import { Injectable } from '@nestjs/common';
import { InjectRepository } from '@nestjs/typeorm';
import { Repository, In } from 'typeorm';
import * as jwt from 'jsonwebtoken';

import { FactionEntity } from '../faction/faction.entity';
import { GameEntity } from '../game/game.entity';
import { TrackingService } from '../tracking/tracking.service';
import { UserService } from '../user/user.service';
import { FactionService } from '../faction/faction.service';
import { TrackingEntity } from '../tracking/tracking.entity';
import {
  MapDrawingEntity,
  MapDrawingHistoryEntity,
} from '../draw/coordinate.entity';
import { ScoreService } from 'src/score/score.service';
import { ScoreEntity } from 'src/score/score.entity';

@Injectable()
export class ReplayService {
  constructor(
    @InjectRepository(FactionEntity)
    private factionRepository: Repository<FactionEntity>,
    @InjectRepository(GameEntity)
    private gameRepository: Repository<GameEntity>,
    @InjectRepository(TrackingEntity)
    private trackingRepository: Repository<TrackingEntity>,
    @InjectRepository(MapDrawingEntity)
    private mapdrawingRepository: Repository<MapDrawingEntity>,
    @InjectRepository(MapDrawingHistoryEntity)
    private mapHistoryRepository: Repository<MapDrawingHistoryEntity>,
    @InjectRepository(ScoreEntity)
    private scoreRepository: Repository<ScoreEntity>,
    private trackingService: TrackingService,
    private userService: UserService,
    private factionService: FactionService,
    private scoreService: ScoreService,
  ) {}

  /*   async replayData(gameId) {
    let mapDrawingIds = await this.mapdrawingRepository.find({
      where: { gameId: gameId },
      select: ['mapDrawingId'],
  // replay data for Factions
  async replayData(gameId) {
    const replay = await this.factionRepository.find({
      where: { game: gameId },
      relations: ['mapDrawings', 'scores', 'trackdata'],
    });
    let drawings = [];
    await Promise.all(
      mapDrawingIds.map(async mapId => {
        drawings.push(
          await this.mapHistoryRepository.find({
            mapdrawing: mapId.mapDrawingId,
          }),
        );
      }),
    );
    return drawings;
  } */

  async replayData(gameId) {
    //
    // this block returns game's initial location
    //
    let gameObj = await this.gameRepository.findOne({
      where: { id: gameId },
      select: ['center'],
    });
    let gamelocation = [gameObj.center.lat, gameObj.center.lng];
    //
    // this block returns all player data from the game
    //
    let playerdata = await this.trackingRepository.find({
      where: { game: gameId },
      relations: ['faction', 'gamepersonId', 'gamepersonId.person'],
    });
    // parse data
    const currentdata = await Promise.all(
      playerdata.map(async player => {
        player.data.map(async data => {
          data['info'] = [
            { key: 'icon', value: player.icon },
            { key: 'colour', value: player.faction.colour },
            { key: 'name', value: player.gamepersonId.person.name },
            { key: 'role', value: player.gamepersonId.role },
          ];
        });
        return player['data'];
      }),
    );
    //
    // this block returns all faction data from the game
    //
    let factions = await this.factionRepository.find({ game: gameId });
    let currentFactions = factions.map(faction => {
      return {
        name: faction.factionName,
        colour: faction.colour,
        // all factions will be rendered on the map on default
        active: true,
      };
    });
    let factionIds = factions.map(faction => faction.factionId);
    //
    // this block returns all score data from the game
    //
    let currentScore = [];
    await Promise.all(
      factions.map(async faction => {
        let scores = await this.scoreRepository.find({
          where: { faction: faction },
          relations: ['faction'],
        });
        currentScore.push(
          scores.map(score => {
            return {
              score: score.score,
              timestamp: score.scoreTimeStamp,
              faction: score.faction['factionName'],
            };
          }),
        );
      }),
    );
    //
    // this block returns all map drawings from the game
    //
    let refs = await this.mapdrawingRepository.find({
      where: { gameId: gameId },
      select: ['mapDrawingId'],
    });

    let drawData = await Promise.all(
      refs.map(async ref => {
        return await this.mapHistoryRepository.find({
          where: { mapdrawing: ref.mapDrawingId },
        });
      }),
    );

    return {
      location: gamelocation,
      players: currentdata,
      factions: currentFactions,
      scores: currentScore,
      drawings: drawData,
    };
  }
  // generate mockdata for a 3 day game
  // create x amount of players
  // assign them to two separate factions
  // assign them to three separate groups
  // insert x amount of locations for each players around some lat lng area
  // insert x amount of score ticks for score mockdata
  // use the game's initial geojson to draw game area
  //
  async mockdata(gameId) {
    // initial settings for mockdata
    // set the x amount of users to be created
    const USER_AMOUNT = 100;
    // set the x amount of locations to be created
    const USER_LOCATIONS = 10;
    // set the LAT and LNG for initial location
    const LAT = 62.24147;
    const LNG = 25.72088;
    // set the score tick amount
    // not used at the moment
    const SCORE_TICKS = 10;
    // setup the arrays for users, gamepersons and groups
    const users = [];
    const gamepersons = [];
    const groups = [];
    // characters for generating random usernames and the length of the username
    const chars =
      'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789';
    const N = 15;
    // used to equally divide the players to factions
    let f = 0;

    // get game info and factions associated with it
    const game = await this.gameRepository.findOne({
      where: { id: gameId },
      relations: ['factions'],
    });
    // get groups for all factions
    await game.factions.forEach(async faction => {
      groups.push(await this.factionService.showGroups(faction.factionId));
    });
    // create x amount of users for the mock game with random username
    for (let i = 0; i < USER_AMOUNT; i++) {
      let res = await this.userService.register({
        name: Array(N)
          .join()
          .split(',')
          .map(function() {
            return chars.charAt(Math.floor(Math.random() * chars.length));
          })
          .join(''),
        password: 'asd',
      });
      // get user information by verifying the token returned by register
      let user = await jwt.verify(res.token, process.env.SECRET);
      // push the created user to users array
      users.push(user);

      // reset index if it's out of range
      if (f === game.factions.length) f = 0;
      // join faction with the created user
      let gameperson = await this.factionService.joinFaction(user['id'], {
        factionId: game.factions[f].factionId,
        factionPassword: game.factions[f].factionPassword,
        game: gameId,
      });
      // join a group randomly
      await this.factionService.joinGroup(gameperson, {
        groupId: groups[f][Math.floor(Math.random() * 3)],
      });
      // push the gameperson ref to gamepersons array
      gamepersons.push(gameperson);
      f++;
    }
    // push x amount of user locations
    for (let i = 1; i < USER_LOCATIONS + 1; i++) {
      let date: number = Date.now();
      let x = 1;
      // score ticks with players to sync them
      await this.scoreService.scoreTick(gameId);
      // add location entry for each gameperson
      await Promise.all(
        gamepersons.map(async gameperson => {
          // format player locations with x-modifier that same factions are closer to each other
          if (x > game.factions.length) x = 1;
          let dataObject = {
            lat: LAT + ((i + Math.random()) * 5) / 2000,
            lng: LNG + x / 100 + (Math.random() * 5) / 2000,
            time: date,
          };
          x++;
          // push the tracking data
          await this.trackingService.trackLocation(
            gameperson,
            gameId,
            dataObject,
          );
        }),
      );
    }

    return {
      message: 'all done',
    };
  }
}
